;-------------------------------------------------------------------------
; Copyright (c)1999 Stephen Freyder, David Helland, and Bruce D. Lightner.
; Use subject to licensing restrictions described in "PICOWEB SERVER
; SOFTWARE LICENSE AGREEMENT DATED JULY 14, 1999" (See file LICENSE.txt).
;-------------------------------------------------------------------------
;+
; **-bootp-bootp support routines.
;
; this module implements the BOOTP protocol, as described in RFC951.
;
; the basic idea is to send a UDP broadcast to 255.255.255.255:bootp, and
; expect a reply TO 255.255.255.255:68 (68 appears to be a canonical request
; port).  in the request PAYLOAD is our own ethernet address, our current
; IP address (even though it might be bogus), and some control flags.
;
; when the reply comes back, we check to make sure it is to port 68, and
; that the ethernet address (again, in the PAYLOAD) matches ours, and also
; that the operation type is "reply".  if all of these tests pass, the
; packet is accepted as valid, and our IP address is changed to the one
; included in the request packet.
;-

#include "bootp.h"

#ifndef NET_CONFIG_IP
.text
    .global bootp
bootp:
    pbegin
    pi2x 0,0xffff
    pi2x 2,0xffff
    pi2x 4,0xffff                           ; dest ether = FF:FF:FF:FF:FF:FF

    SET_IP(pi2x,IP_SRCADDR,0,0,0,1)         ; 0.0.0.1
    pi2x UDP_SP,BSWAP(BOOTP_REPLY_PORT)     ;  :reply_port

    SET_IP(pi2x,IP_DSTADDR,255,255,255,255) ; 255.255.255.255
    pi2x UDP_DP,BSWAP(BOOTP_REQUEST_PORT)   ;  :bootp

    pi2x IP_LEN,BSWAP(IP_HEADER_LEN+UDP_HEADER_LEN+BOOTP_LEN)
    pi2x UDP_LEN,BSWAP(UDP_HEADER_LEN+BOOTP_LEN)

    pz2x BOOTP_OP,BOOTP_LEN         ; zero UDP data portion of xmit buffer

    pi2x BOOTP_OP,0x0101            ; optype=01 (req), hwtype=01 (ether)
    pi2x BOOTP_HLEN,0x0006          ; addrlen=06 (ether), hops=00
    ps2x BOOTP_CHADDR,my_ether,6    ; insert my ether address
    ps2x BOOTP_CIADDR,my_ip,4       ;  and my IP address (even if wrong)
    pcall ip_header_1               ; do header except ip addresses
    pcall udp_checksum
    px2s buf,IP_LEN,2               ; read back the length
    pmovwi buf,[swap buf]           ; swap bytes
    paddwi buf,6+6+2                ; compute length with ether header
    xmit_frame [buf]                ; out she goes

    pjump exit_now

;+
; **-bootp_timer-timer expiration.
;
; this routine is called whenever the bootp timer expires.  if we do not know
; our IP address yet, then generate another bootp request.
;-
.text
.global bootp_timer
bootp_timer:
    lds r0,my_ip                    ; get first byte of IP address
    or r0,r0
    brne bootp_bail
    rjmp bootp                      ; generate a bootp request
bootp_bail:
    ret

#endif /* !NET_CONFIG_IP */

;+
; **-bootp_reply-reply to bootp request from net
;
; this routine is called whenever a bootp reply packet is
; received - send a reply if it is for us and opcode is OK.
; if NET_CONFIG_IP is defined, also save address in (S)EEPROM.
;-

#ifdef NET_CONFIG_IP
.section seeprom
#else
.text
#endif

bootp_reply:
    pr2s buf,BOOTP_CHADDR,6         ; get the ethernet address
    pcmpn buf,my_ether,6            ; see if it is my ether address
    pjumpne exit_now                ; nope - it stinks
    pr2s buf,BOOTP_OP,2             ; get optype and hwtype
    pcmpwi buf,0x0102               ; BOOTREPLY(02), ETHER(01)
    pjumpne exit_now                ; sorry - not a good reply packet

#ifdef NET_CONFIG_IP
    poksetip                        ; is it safe to change IP via net?
    pjumpne exit_now                ; sorry - not allowed!
    pr2s buf,BOOTP_YIADDR,4         ; get reply IP into temp RAM
    pr2s buf+4,BOOTP_VEND,8         ; get reply NM/GW into temp RAM
    pcmpn buf,my_ip,12              ; has IP, NM or GW changed? (12 bytes!)
    pjumpeq exit_now                ; quick exit if still the same
#endif

#if defined(BOOTP_DEBUG)
    pprintr "IP",BOOTP_YIADDR,4
    pcrlf
#endif

    pr2s my_ip,BOOTP_YIADDR,4       ; change my IP address (in RAM)

#ifdef NET_CONFIG_IP
    ps2ee eeprom_ip_addr,buf,12     ; change IP, NM and GW in EEPROM
    pjump pexit_reset
#else /* NET_CONFIG_IP */
    pjump exit_now
#endif /* !NET_CONFIG_IP */

;+
; **-unknown-ip-unknown ip address packet handler.
;
; called from network when packet is not for us - check for bootp
; reply packet.
;-

.text
    .global import_unknown_ipaddr
import_unknown_ipaddr:
    pbegin
#if 0
  pserial_debug
  pprint "import_unknown_ipaddr\r\n";
#endif

    ;;pcmpwi buf,0xffff               ; 255.255
    ;;pjumpne exit_now                ; not broadcast
    ;;pcmpwi buf+2,0xffff             ; .255.255
    ;;pjumpne exit_now                ; not broadcast
;
; OK - it is to the correct address, how about the proto - is it UDP? 
;
    pr2s buf,IPH+8,2                ; get TTL (buf) and proto (buf+1)
    pcmpbi buf+1,0x11               ; check if UDP
    pjumpne exit_now                ; sorry - not UDP
;
; UDP, check the port.
;
    pr2s buf,UDP_DP,2               ; get destination port
    pcmpwi buf,BSWAP(BOOTP_REPLY_PORT)      ; check if to bootp port
    pjumpeq bootp_reply             ; smells good - act on it

exit_now:
    pend
    ret

#ifdef NET_CONFIG_IP
pexit_reset:
    pend
    rjmp RESET                      ; reset processor (will use new IP)
#endif /* NET_CONFIG_IP */
